@@ -19,16 +19,22 @@ package hostpathprovisioner
 import (
 	"context"
 	"fmt"
+	"reflect"
+	"time"
 
+	promv1 "github.com/coreos/prometheus-operator/pkg/apis/monitoring/v1"
 	"github.com/go-logr/logr"
 	secv1 "github.com/openshift/api/security/v1"
 	conditions "github.com/openshift/custom-resource-status/conditions/v1"
 	"github.com/operator-framework/operator-sdk/pkg/k8sutil"
+	"github.com/prometheus/client_golang/prometheus"
 	appsv1 "k8s.io/api/apps/v1"
 	corev1 "k8s.io/api/core/v1"
 	rbacv1 "k8s.io/api/rbac/v1"
+	storagev1 "k8s.io/api/storage/v1"
 	"k8s.io/apimachinery/pkg/api/errors"
 	"k8s.io/apimachinery/pkg/api/meta"
+	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
 	"k8s.io/apimachinery/pkg/runtime"
 	"k8s.io/apimachinery/pkg/types"
 	"k8s.io/client-go/tools/record"
@@ -40,18 +46,39 @@ import (
 	"sigs.k8s.io/controller-runtime/pkg/handler"
 	logf "sigs.k8s.io/controller-runtime/pkg/log"
 	"sigs.k8s.io/controller-runtime/pkg/manager"
+	"sigs.k8s.io/controller-runtime/pkg/metrics"
 	"sigs.k8s.io/controller-runtime/pkg/reconcile"
 	"sigs.k8s.io/controller-runtime/pkg/source"
 )
 
-var log = logf.Log.WithName("controller_hostpathprovisioner")
-var watchNamespaceFunc = k8sutil.GetWatchNamespace
+var (
+	log                = logf.Log.WithName("controller_hostpathprovisioner")
+	watchNamespaceFunc = k8sutil.GetWatchNamespace
+	readyGauge         = prometheus.NewGauge(
+		prometheus.GaugeOpts{
+			Name: "kubevirt_hpp_cr_ready",
+			Help: "HPP CR Ready",
+		})
+)
+
+func init() {
+	metrics.Registry = prometheus.NewRegistry()
+	metrics.Registry.MustRegister(readyGauge)
+	// 0 is our 'something bad is going on' value for alert to start firing, so can't default to that
+	readyGauge.Set(-1)
+}
+
+const (
+	snapshotFeatureGate = "Snapshotting"
+	hppFinalizer        = "finalizer.delete.hostpath-provisioner"
+)
 
 func isErrCacheNotStarted(err error) bool {
 	if err == nil {
 		return false
 	}
-	return err.(*cache.ErrCacheNotStarted) != nil
+	_, ok := err.(*cache.ErrCacheNotStarted)
+	return ok
 }
 
 // Add creates a new HostPathProvisioner Controller and adds it to the Manager. The Manager will set fields on the Controller
@@ -84,8 +111,8 @@ func add(mgr manager.Manager, r reconcile.Reconciler) error {
 	}
 
 	// mapFn will be used to map reconcile requests to the HPP for resources that don't have an ownerRef
-	mapFn := handler.ToRequestsFunc(func(o handler.MapObject) []reconcile.Request {
-		if val, ok := o.Meta.GetLabels()["k8s-app"]; ok && val == MultiPurposeHostPathProvisionerName {
+	mapFn := handler.MapFunc(func(o client.Object) []reconcile.Request {
+		if val, ok := o.GetLabels()["k8s-app"]; ok && val == MultiPurposeHostPathProvisionerName {
 			hppList, err := getHppList(mgr.GetClient())
 			if err != nil {
 				log.Error(err, "Error getting HPPs")
@@ -117,6 +144,14 @@ func add(mgr manager.Manager, r reconcile.Reconciler) error {
 		return err
 	}
 
+	err = c.Watch(&source.Kind{Type: &appsv1.Deployment{}}, &handler.EnqueueRequestForOwner{
+		IsController: true,
+		OwnerType:    &hostpathprovisionerv1.HostPathProvisioner{},
+	})
+	if err != nil {
+		return err
+	}
+
 	err = c.Watch(&source.Kind{Type: &corev1.ServiceAccount{}}, &handler.EnqueueRequestForOwner{
 		IsController: true,
 		OwnerType:    &hostpathprovisionerv1.HostPathProvisioner{},
@@ -125,25 +160,36 @@ func add(mgr manager.Manager, r reconcile.Reconciler) error {
 		return err
 	}
 
-	err = c.Watch(&source.Kind{Type: &rbacv1.ClusterRoleBinding{}}, &handler.EnqueueRequestsFromMapFunc{
-		ToRequests: mapFn,
+	err = c.Watch(&source.Kind{Type: &rbacv1.RoleBinding{}}, &handler.EnqueueRequestForOwner{
+		IsController: true,
+		OwnerType:    &hostpathprovisionerv1.HostPathProvisioner{},
 	})
 	if err != nil {
 		return err
 	}
 
-	err = c.Watch(&source.Kind{Type: &rbacv1.ClusterRole{}}, &handler.EnqueueRequestsFromMapFunc{
-		ToRequests: mapFn,
+	err = c.Watch(&source.Kind{Type: &rbacv1.Role{}}, &handler.EnqueueRequestForOwner{
+		IsController: true,
+		OwnerType:    &hostpathprovisionerv1.HostPathProvisioner{},
 	})
 	if err != nil {
 		return err
 	}
 
-	if used, err := r.(*ReconcileHostPathProvisioner).checkSCCUsed(); used == true || isErrCacheNotStarted(err) {
-		err = c.Watch(&source.Kind{Type: &secv1.SecurityContextConstraints{}}, &handler.EnqueueRequestsFromMapFunc{
-			ToRequests: mapFn,
-		})
-		if err != nil {
+	if err := c.Watch(&source.Kind{Type: &storagev1.CSIDriver{}}, handler.EnqueueRequestsFromMapFunc(mapFn)); err != nil {
+		return err
+	}
+
+	if err := c.Watch(&source.Kind{Type: &rbacv1.ClusterRoleBinding{}}, handler.EnqueueRequestsFromMapFunc(mapFn)); err != nil {
+		return err
+	}
+
+	if err := c.Watch(&source.Kind{Type: &rbacv1.ClusterRole{}}, handler.EnqueueRequestsFromMapFunc(mapFn)); err != nil {
+		return err
+	}
+
+	if used, err := r.(*ReconcileHostPathProvisioner).checkSCCUsed(); used || isErrCacheNotStarted(err) {
+		if err := c.Watch(&source.Kind{Type: &secv1.SecurityContextConstraints{}}, handler.EnqueueRequestsFromMapFunc(mapFn)); err != nil {
 			if meta.IsNoMatchError(err) {
 				log.Info("Not watching SecurityContextConstraints")
 				return nil
@@ -152,6 +198,32 @@ func add(mgr manager.Manager, r reconcile.Reconciler) error {
 		}
 	}
 
+	if used, err := r.(*ReconcileHostPathProvisioner).checkPrometheusUsed(); used || isErrCacheNotStarted(err) {
+		if err := c.Watch(&source.Kind{Type: &promv1.PrometheusRule{}}, handler.EnqueueRequestsFromMapFunc(mapFn)); err != nil {
+			if meta.IsNoMatchError(err) {
+				log.Info("Not watching PrometheusRules")
+				return nil
+			}
+			return err
+		}
+		if err := c.Watch(&source.Kind{Type: &promv1.ServiceMonitor{}}, handler.EnqueueRequestsFromMapFunc(mapFn)); err != nil {
+			if meta.IsNoMatchError(err) {
+				log.Info("Not watching ServiceMonitors")
+				return nil
+			}
+			return err
+		}
+		if err := c.Watch(&source.Kind{Type: &rbacv1.Role{}}, handler.EnqueueRequestsFromMapFunc(mapFn)); err != nil {
+			return err
+		}
+		if err := c.Watch(&source.Kind{Type: &rbacv1.RoleBinding{}}, handler.EnqueueRequestsFromMapFunc(mapFn)); err != nil {
+			return err
+		}
+		if err := c.Watch(&source.Kind{Type: &corev1.Service{}}, handler.EnqueueRequestsFromMapFunc(mapFn)); err != nil {
+			return err
+		}
+	}
+
 	return nil
 }
 
@@ -172,7 +244,7 @@ type ReconcileHostPathProvisioner struct {
 // and what is in the HostPathProvisioner.Spec
 // The Controller will requeue the Request to be processed again if the returned error is non-nil or
 // Result.Requeue is true, otherwise upon completion it will remove the work from the queue.
-func (r *ReconcileHostPathProvisioner) Reconcile(request reconcile.Request) (reconcile.Result, error) {
+func (r *ReconcileHostPathProvisioner) Reconcile(context context.Context, request reconcile.Request) (reconcile.Result, error) {
 	reqLogger := r.Log.WithValues("Request.Namespace", request.Namespace, "Request.Name", request.Name)
 	reqLogger.V(3).Info("Reconciling HostPathProvisioner")
 
@@ -183,7 +255,7 @@ func (r *ReconcileHostPathProvisioner) Reconcile(request reconcile.Request) (rec
 		return reconcile.Result{}, err
 	}
 	if size := len(hppList.Items); size > 1 {
-		err := fmt.Errorf("There should be a single hostpath provisioner, %d items found", size)
+		err := fmt.Errorf("there should be a single hostpath provisioner, %d items found", size)
 		reqLogger.Error(err, "Multiple HPPs detected")
 		return reconcile.Result{}, err
 	}
@@ -195,7 +267,7 @@ func (r *ReconcileHostPathProvisioner) Reconcile(request reconcile.Request) (rec
 
 	// Fetch the HostPathProvisioner instance
 	cr := &hostpathprovisionerv1.HostPathProvisioner{}
-	err = r.client.Get(context.TODO(), request.NamespacedName, cr)
+	err = r.client.Get(context, request.NamespacedName, cr)
 	if err != nil {
 		if errors.IsNotFound(err) {
 			// Request object not found, could have been deleted after reconcile request.
@@ -206,9 +278,52 @@ func (r *ReconcileHostPathProvisioner) Reconcile(request reconcile.Request) (rec
 		// Error reading the object - requeue the request.
 		return reconcile.Result{}, err
 	}
+	if r.isLegacy(cr) {
+		reqLogger.Info("Detected legacy CR, Reconciling CSI and legacy controller plugin")
+	} else {
+		reqLogger.Info("Detected CSI CR, reconciling CSI only")
+	}
 
-	isMarkedToBeDeleted := cr.GetDeletionTimestamp() != nil
-	if isMarkedToBeDeleted {
+	// Ready metric so we can alert whenever we are not ready for a while
+	if IsHppAvailable(cr) {
+		readyGauge.Set(1)
+	} else if !IsHppProgressing(cr) {
+		// Not an issue if progress is still ongoing
+		readyGauge.Set(0)
+	}
+
+	namespace, err := watchNamespaceFunc()
+	if err != nil {
+		MarkCrFailed(cr, watchNameSpace, err.Error())
+		r.recorder.Event(cr, corev1.EventTypeWarning, watchNameSpace, err.Error())
+		err2 := r.client.Update(context, cr)
+		if err2 != nil {
+			reqLogger.Error(err2, "Unable to update CR to failed state")
+		}
+		// Error reading the object - requeue the request.
+		return reconcile.Result{}, err
+	}
+
+	if cr.GetDeletionTimestamp() != nil {
+		if err := r.cleanDeployments(reqLogger, cr, namespace); err != nil {
+			return reconcile.Result{}, err
+		}
+		deployments, err := r.currentStoragePoolDeployments(reqLogger, cr, namespace)
+		if err != nil {
+			return reconcile.Result{}, err
+		}
+		reqLogger.Info("Number of deployments still active", "count", len(deployments))
+		cleanupFinished, err := r.hasCleanUpFinished()
+		if err != nil {
+			return reconcile.Result{}, err
+		}
+		if len(deployments) == 0 && cleanupFinished {
+			if err := r.removeCleanUpJobs(reqLogger); err != nil {
+				return reconcile.Result{}, err
+			}
+		} else {
+			return reconcile.Result{RequeueAfter: time.Second}, nil
+		}
 		reqLogger.Info("Deleting SecurityContextConstraint", "SecurityContextConstraints", MultiPurposeHostPathProvisionerName)
 		if err := r.deleteSCC(MultiPurposeHostPathProvisionerName); err != nil {
 			reqLogger.Error(err, "Unable to delete SecurityContextConstraints")
@@ -216,21 +331,28 @@ func (r *ReconcileHostPathProvisioner) Reconcile(request reconcile.Request) (rec
 			// should be not return and allow the CR to be deleted but without deleting the SCC if that fails.
 			return reconcile.Result{}, err
 		}
-		reqLogger.Info("Deleting ClusterRoleBinding", "ClusterRoleBinding", MultiPurposeHostPathProvisionerName)
-		if err := r.deleteClusterRoleBindingObject(MultiPurposeHostPathProvisionerName); err != nil {
-			reqLogger.Error(err, "Unable to delete ClusterRoleBinding")
+		if err := r.deleteSCC(fmt.Sprintf("%s-csi", MultiPurposeHostPathProvisionerName)); err != nil {
+			reqLogger.Error(err, "Unable to delete CSI SecurityContextConstraints")
+			// TODO, should we return and in essence keep retrying, and thus never be able to delete the CR if deleting the SCC fails, or
+			// should be not return and allow the CR to be deleted but without deleting the SCC if that fails.
 			return reconcile.Result{}, err
 		}
-		reqLogger.Info("Deleting ClusterRole", "ClusterRole", MultiPurposeHostPathProvisionerName)
-		if err := r.deleteClusterRoleObject(MultiPurposeHostPathProvisionerName); err != nil {
-			reqLogger.Error(err, "Unable to delete ClusterRole")
+		if err := r.deletePrometheusResources(namespace); err != nil {
+			reqLogger.Error(err, "Unable to delete Prometheus Infra (PrometheusRule, ServiceMonitor, RBAC)")
 			return reconcile.Result{}, err
 		}
-
-		cr.SetFinalizers(nil)
+		if res, err := r.deleteAllRbac(reqLogger, namespace); err != nil {
+			return res, err
+		}
+		reqLogger.Info("Deleting CSIDriver", "CSIDriver", MultiPurposeHostPathProvisionerName)
+		if err := r.deleteCSIDriver(); err != nil {
+			reqLogger.Error(err, "Unable to delete CSIDriver")
+			return reconcile.Result{}, err
+		}
+		RemoveFinalizer(cr, hppFinalizer)
 
 		// Update CR
-		err = r.client.Update(context.TODO(), cr)
+		err = r.client.Update(context, cr)
 		if err != nil {
 			reqLogger.Error(err, "Unable to remove finalizer from CR")
 			return reconcile.Result{}, err
@@ -238,23 +360,12 @@ func (r *ReconcileHostPathProvisioner) Reconcile(request reconcile.Request) (rec
 		return reconcile.Result{}, nil
 	}
 
+	currentCopy := cr.DeepCopy()
 	// Add finalizer for this CR
 	if err := r.addFinalizer(reqLogger, cr); err != nil {
 		return reconcile.Result{}, err
 	}
 
-	namespace, err := watchNamespaceFunc()
-	if err != nil {
-		MarkCrFailed(cr, watchNameSpace, err.Error())
-		r.recorder.Event(cr, corev1.EventTypeWarning, watchNameSpace, err.Error())
-		err2 := r.client.Update(context.TODO(), cr)
-		if err2 != nil {
-			reqLogger.Error(err2, "Unable to update CR to failed state")
-		}
-		// Error reading the object - requeue the request.
-		return reconcile.Result{}, err
-	}
-
 	cr.Status.OperatorVersion = versionString
 	cr.Status.TargetVersion = versionString
 	canUpgrade, err := canUpgrade(cr.Status.ObservedVersion, versionString)
@@ -266,7 +377,7 @@ func (r *ReconcileHostPathProvisioner) Reconcile(request reconcile.Request) (rec
 		//New install, mark deploying.
 		MarkCrDeploying(cr, deployStarted, deployStartedMessage)
 		r.recorder.Event(cr, corev1.EventTypeNormal, deployStarted, deployStartedMessage)
-		err = r.client.Update(context.TODO(), cr)
+		err = r.client.Update(context, cr)
 		if err != nil {
 			reqLogger.Info("Marked deploying failed", "Error", err.Error())
 			// Error updating the object - requeue the request.
@@ -279,7 +390,7 @@ func (r *ReconcileHostPathProvisioner) Reconcile(request reconcile.Request) (rec
 		MarkCrUpgradeHealingDegraded(cr, upgradeStarted, fmt.Sprintf("Started upgrade to version %s", cr.Status.TargetVersion))
 		r.recorder.Event(cr, corev1.EventTypeWarning, upgradeStarted, fmt.Sprintf("Started upgrade to version %s", cr.Status.TargetVersion))
 		// Mark Observed version to blank, so we get to the reconcile upgrade section.
-		err = r.client.Update(context.TODO(), cr)
+		err = r.client.Update(context, cr)
 		if err != nil {
 			// Error updating the object - requeue the request.
 			return reconcile.Result{}, err
@@ -289,26 +400,79 @@ func (r *ReconcileHostPathProvisioner) Reconcile(request reconcile.Request) (rec
 
 	res, err := r.reconcileUpdate(reqLogger, request, cr, namespace)
 	if err == nil {
-		// Check if all requested pods are available.
-		degraded, err := r.checkDegraded(reqLogger, cr, namespace)
-		if err != nil {
-			return reconcile.Result{}, err
-		}
-		if !degraded && cr.Status.ObservedVersion != versionString {
-			cr.Status.ObservedVersion = versionString
-			err = r.client.Update(context.TODO(), cr)
-			if err != nil {
-				// Error updating the object - requeue the request.
-				return reconcile.Result{}, err
-			}
-		}
+		res, err = r.reconcileStatus(context, reqLogger, cr, namespace, versionString)
 	} else {
 		MarkCrFailedHealing(cr, reconcileFailed, fmt.Sprintf("Unable to successfully reconcile: %v", err))
 		r.recorder.Event(cr, corev1.EventTypeWarning, reconcileFailed, fmt.Sprintf("Unable to successfully reconcile: %v", err))
 	}
+
+	r.ignoreHeartBeatTimestamp(currentCopy, cr)
+	if !reflect.DeepEqual(currentCopy, cr) {
+		logJSONDiff(reqLogger, currentCopy, cr)
+		updateErr := r.client.Update(context, cr)
+		if updateErr != nil {
+			r.Log.Error(err, "Unable to successfully reconcile")
+			err = updateErr
+		}
+	}
 	return res, err
 }
 
+func (r *ReconcileHostPathProvisioner) ignoreHeartBeatTimestamp(currentCopy, cr *hostpathprovisionerv1.HostPathProvisioner) {
+	for i, condition := range currentCopy.Status.Conditions {
+		crCond := conditions.FindStatusCondition(cr.Status.Conditions, condition.Type)
+		if crCond.Message == condition.Message && crCond.Reason == condition.Reason && crCond.Status == condition.Status {
+			currentCopy.Status.Conditions[i].LastHeartbeatTime = crCond.LastHeartbeatTime
+		}
+	}
+}
+
+func (r *ReconcileHostPathProvisioner) isLegacy(cr *hostpathprovisionerv1.HostPathProvisioner) bool {
+	return cr.Spec.PathConfig != nil
+}
+
+func (r *ReconcileHostPathProvisioner) reconcileStatus(context context.Context, reqLogger logr.Logger, cr *hostpathprovisionerv1.HostPathProvisioner, namespace, versionString string) (reconcile.Result, error) {
+	// Check if all requested pods are available.
+	degraded, err := r.checkDegraded(reqLogger, cr, namespace)
+	if err != nil {
+		return reconcile.Result{}, err
+	}
+	if err := r.reconcileStoragePoolStatus(reqLogger, cr, namespace); err != nil {
+		MarkCrFailedHealing(cr, "StoragePoolNotReady", err.Error())
+		return reconcile.Result{}, err
+	}
+	if !degraded && cr.Status.ObservedVersion != versionString {
+		cr.Status.ObservedVersion = versionString
+	}
+	return reconcile.Result{}, nil
+}
+
+func (r *ReconcileHostPathProvisioner) deleteAllRbac(reqLogger logr.Logger, namespace string) (reconcile.Result, error) {
+	for _, name := range []string{ProvisionerServiceAccountName, ProvisionerServiceAccountNameCsi, MultiPurposeHostPathProvisionerName} {
+		reqLogger.Info("Deleting ClusterRoleBinding", "ClusterRoleBinding", name)
+		if err := r.deleteClusterRoleBindingObject(name); err != nil {
+			reqLogger.Error(err, "Unable to delete ClusterRoleBinding")
+			return reconcile.Result{}, err
+		}
+		reqLogger.Info("Deleting ClusterRole", "ClusterRole", name)
+		if err := r.deleteClusterRoleObject(name); err != nil {
+			reqLogger.Error(err, "Unable to delete ClusterRole")
+			return reconcile.Result{}, err
+		}
+		reqLogger.Info("Deleting RoleBinding", "ClusterRoleBinding", name)
+		if err := r.deleteRoleBindingObject(name, namespace); err != nil {
+			reqLogger.Error(err, "Unable to delete RoleBinding")
+			return reconcile.Result{}, err
+		}
+		reqLogger.Info("Deleting Role", "ClusterRole", name)
+		if err := r.deleteRoleObject(name, namespace); err != nil {
+			reqLogger.Error(err, "Unable to delete Role")
+			return reconcile.Result{}, err
+		}
+	}
+	return reconcile.Result{}, nil
+}
+
 func canUpgrade(current, target string) (bool, error) {
 	if current == "" {
 		// Can't upgrade if no current is set
@@ -337,41 +501,70 @@ func canUpgrade(current, target string) (bool, error) {
 
 func (r *ReconcileHostPathProvisioner) reconcileUpdate(reqLogger logr.Logger, request reconcile.Request, cr *hostpathprovisionerv1.HostPathProvisioner, namespace string) (reconcile.Result, error) {
 	// Reconcile the objects this operator manages.
-	res, err := r.reconcileDaemonSet(reqLogger, cr, namespace, r.recorder)
+	res, err := r.reconcileDaemonSet(reqLogger, cr, namespace)
 	if err != nil {
-		reqLogger.Error(err, "Unable to create DaemonSet")
-		return reconcile.Result{}, err
+		reqLogger.Error(err, "unable to create DaemonSet")
+		return res, err
+	}
+	// Reconcile storage pools
+	res, err = r.reconcileStoragePools(reqLogger, cr, namespace)
+	if err != nil {
+		reqLogger.Error(err, "unable to configure storage pools")
+		return res, err
 	}
 	res, err = r.reconcileServiceAccount(reqLogger, cr, namespace)
 	if err != nil {
-		reqLogger.Error(err, "Unable to create ServiceAccount")
-		return reconcile.Result{}, err
+		reqLogger.Error(err, "unable to create ServiceAccount")
+		return res, err
 	}
-	res, err = r.reconcileClusterRole(reqLogger, cr, r.recorder)
+	res, err = r.reconcileClusterRole(reqLogger, cr)
 	if err != nil {
-		reqLogger.Error(err, "Unable to create ClusterRole")
-		return reconcile.Result{}, err
+		reqLogger.Error(err, "unable to create ClusterRole")
+		return res, err
 	}
-	res, err = r.reconcileClusterRoleBinding(reqLogger, cr, namespace, r.recorder)
+	res, err = r.reconcileClusterRoleBinding(reqLogger, cr, namespace)
 	if err != nil {
-		reqLogger.Error(err, "Unable to create ClusterRoleBinding")
-		return reconcile.Result{}, err
+		reqLogger.Error(err, "unable to create ClusterRoleBinding")
+		return res, err
 	}
-	res, err = r.reconcileSecurityContextConstraints(reqLogger, cr, namespace, r.recorder)
+	res, err = r.reconcileRole(reqLogger, cr, namespace)
 	if err != nil {
-		reqLogger.Error(err, "Unable to create SecurityContextConstraints")
-		return reconcile.Result{}, err
+		reqLogger.Error(err, "unable to create Role")
+		return res, err
+	}
+	res, err = r.reconcileRoleBinding(reqLogger, cr, namespace)
+	if err != nil {
+		reqLogger.Error(err, "unable to create RoleBinding")
+		return res, err
+	}
+	res, err = r.reconcileCSIDriver(reqLogger, cr, namespace)
+	if err != nil {
+		reqLogger.Error(err, "unable to create CSIDriver")
+		return res, err
+	}
+	res, err = r.reconcileSecurityContextConstraints(reqLogger, cr, namespace)
+	if err != nil {
+		reqLogger.Error(err, "unable to create SecurityContextConstraints")
+		return res, err
+	}
+	res, err = r.reconcilePrometheusInfra(reqLogger, cr, namespace)
+	if err != nil {
+		reqLogger.Error(err, "unable to create Prometheus Infra (PrometheusRule, ServiceMonitor, RBAC)")
+		return res, err
 	}
 	daemonSet := &appsv1.DaemonSet{}
-	err = r.client.Get(context.TODO(), types.NamespacedName{Name: MultiPurposeHostPathProvisionerName, Namespace: namespace}, daemonSet)
-	if err != nil {
+	if r.isLegacy(cr) {
+		if err := r.client.Get(context.TODO(), types.NamespacedName{Name: MultiPurposeHostPathProvisionerName, Namespace: namespace}, daemonSet); err != nil {
+			return reconcile.Result{}, err
+		}
+	}
+	daemonSetCsi := &appsv1.DaemonSet{}
+	if err := r.client.Get(context.TODO(), types.NamespacedName{Name: fmt.Sprintf("%s-csi", MultiPurposeHostPathProvisionerName), Namespace: namespace}, daemonSetCsi); err != nil {
 		return reconcile.Result{}, err
 	}
-	if checkApplicationAvailable(daemonSet) {
-		if !IsCrHealthy(cr) {
-			r.recorder.Event(cr, corev1.EventTypeNormal, provisionerHealthy, provisionerHealthyMessage)
-		}
+	if (!r.isLegacy(cr) || checkDaemonSetReady(daemonSet)) && checkDaemonSetReady(daemonSetCsi) {
 		MarkCrHealthyMessage(cr, "Complete", "Application Available")
+		r.recorder.Event(cr, corev1.EventTypeNormal, provisionerHealthy, provisionerHealthyMessage)
 	}
 	return res, nil
 }
@@ -380,12 +573,19 @@ func (r *ReconcileHostPathProvisioner) checkDegraded(logger logr.Logger, cr *hos
 	degraded := false
 
 	daemonSet := &appsv1.DaemonSet{}
-	err := r.client.Get(context.TODO(), types.NamespacedName{Name: MultiPurposeHostPathProvisionerName, Namespace: namespace}, daemonSet)
+	if r.isLegacy(cr) {
+		err := r.client.Get(context.TODO(), types.NamespacedName{Name: MultiPurposeHostPathProvisionerName, Namespace: namespace}, daemonSet)
+		if err != nil {
+			return true, err
+		}
+	}
+	daemonSetCsi := &appsv1.DaemonSet{}
+	err := r.client.Get(context.TODO(), types.NamespacedName{Name: fmt.Sprintf("%s-csi", MultiPurposeHostPathProvisionerName), Namespace: namespace}, daemonSetCsi)
 	if err != nil {
 		return true, err
 	}
 
-	if !checkDaemonSetReady(daemonSet) {
+	if !((!r.isLegacy(cr) || checkDaemonSetReady(daemonSet)) && checkDaemonSetReady(daemonSetCsi)) {
 		degraded = true
 	}
 
@@ -416,21 +616,33 @@ func checkApplicationAvailable(daemonSet *appsv1.DaemonSet) bool {
 	return daemonSet.Status.NumberReady > 0
 }
 
-func (r *ReconcileHostPathProvisioner) addFinalizer(reqLogger logr.Logger, cr *hostpathprovisionerv1.HostPathProvisioner) error {
-	if len(cr.GetFinalizers()) < 1 && cr.GetDeletionTimestamp() == nil {
-		reqLogger.Info("Adding deletion Finalizer")
-		cr.SetFinalizers([]string{"finalizer.delete.hostpath-provisioner"})
-
-		// Update CR
-		err := r.client.Update(context.TODO(), cr)
-		if err != nil {
-			reqLogger.Error(err, "Failed to update cr with finalizer")
-			return err
+func (r *ReconcileHostPathProvisioner) addFinalizer(reqLogger logr.Logger, obj client.Object) error {
+	if obj.GetDeletionTimestamp() == nil {
+		currentFinalizers := obj.GetFinalizers()
+		reqLogger.V(3).Info("Adding deletion Finalizer")
+		AddFinalizer(obj, hppFinalizer)
+		// Only update if we modified the finalizers.
+		if !reflect.DeepEqual(currentFinalizers, obj.GetFinalizers()) {
+			// Update CR
+			err := r.client.Update(context.TODO(), obj)
+			if err != nil {
+				reqLogger.Error(err, "Failed to update cr with finalizer")
+				return err
+			}
 		}
 	}
 	return nil
 }
 
+func (r *ReconcileHostPathProvisioner) isFeatureGateEnabled(feature string, cr *hostpathprovisionerv1.HostPathProvisioner) bool {
+	for _, featuregate := range cr.Spec.FeatureGates {
+		if featuregate == feature {
+			return true
+		}
+	}
+	return false
+}
+
 // This function returns the list of HPP instances in the cluster and an error otherwise
 func getHppList(c client.Client) (*hostpathprovisionerv1.HostPathProvisionerList, error) {
 	hppList := &hostpathprovisionerv1.HostPathProvisionerList{}
@@ -441,3 +653,38 @@ func getHppList(c client.Client) (*hostpathprovisionerv1.HostPathProvisionerList
 
 	return hppList, nil
 }
+
+// AddFinalizer adds a finalizer to a resource
+func AddFinalizer(obj metav1.Object, name string) {
+	if HasFinalizer(obj, name) {
+		return
+	}
+
+	obj.SetFinalizers(append(obj.GetFinalizers(), name))
+}
+
+// RemoveFinalizer removes a finalizer from a resource
+func RemoveFinalizer(obj metav1.Object, name string) {
+	if !HasFinalizer(obj, name) {
+		return
+	}
+
+	var finalizers []string
+	for _, f := range obj.GetFinalizers() {
+		if f != name {
+			finalizers = append(finalizers, f)
+		}
+	}
+
+	obj.SetFinalizers(finalizers)
+}
+
+// HasFinalizer returns true if a resource has a specific finalizer
+func HasFinalizer(object metav1.Object, value string) bool {
+	for _, f := range object.GetFinalizers() {
+		if f == value {
+			return true
+		}
+	}
+	return false
+}
